/*{{{  Header*/
/*------------------------------------------------------------------*/
/*                                                                  */
/*                            socket.c                              */
/*                                                                  */
/*     by Nick Clarke, Southampton University Computing Services    */
/*                                                                  */
/*     Functions added to iserver in order to implement a BSD 4.3   */
/*     socket library on the root transputer.                       */
/*                                                                  */
/*     Tag choices correspond to Inmos B300 (TCPLINK) tags.         */
/*	Socket request packet:					*/
/*	   <FunctionTag:8><SocketOp:8><Args>			*/
/*	Socket reply packet:					*/
/*	   <result:8><SocketErr:32><Signal:32><results>		*/
/*                                                                  */
/*------------------------------------------------------------------*/
/*}}}*/
/*{{{  Includes, Constants and Macros*/
#include <sys/types.h>
#ifdef XCC
#include <sys/size_t.h>
#endif
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <process.h>
#include <pwd.h>
#include <limits.h>
#include <dirent.h>
#include <conc.h>
#include <node.h>
void free(void *);
void raise(int);
char *malloc(int);
#pragma IMS_translate(server_transaction,"server_transaction%c")
extern int server_transaction(unsigned char *, int, unsigned char *);

static enum {
	SpBuffSize=512,
	EnvSize=500,
	NoEnv=4
};

#undef DEBUG
#ifdef DEBUG
#define PRINT_ERROR {\
	char buf[50];\
	sprintf(buf, "%d: err %d\n", __LINE__ , NewErrno);\
	write( 2, buf, strlen(buf)+1);\
}
#else
#define PRINT_ERROR	
#endif
#ifdef XCC
#define SP_O_RDONLY     0x0001
#define SP_O_WRONLY     0x0002
#define SP_O_RDWR       0x0004
#define SP_O_NDELAY     0x0008
#define SP_O_APPEND     0x0010          /* SP_OPEN2 oflags */
#define SP_O_SYNC       0x0020
#define SP_O_CREAT      0x0040
#define SP_O_TRUNC      0x0080
#define SP_O_EXCL       0x0100
#define SP_O_BINARY     0x0200
#define SP_O_TEXT       0x0400

#define _O_RDONLY     0x0
#define _O_WRONLY     0x1
#define _O_RDWR       0x2
#define _O_NDELAY     0x4
#define _O_APPEND     0x8          /* _OPEN2 oflags */
#define _O_SYNC       0x10
#define _O_CREAT      0x100
#define _O_TRUNC      0x200
#define _O_EXCL       0x400

#define SP_SEEK_SET	1
#define SP_SEEK_CUR	2
#define SP_SEEK_END	3
#endif


#define PACK_ITEM(x,y,c) { char *n=(char *)&(x); int i; \
        for(i=0;i<sizeof(x);i++) (y)[c++]=*(n++); }
#define PACK_SLICE(x,l,y,c) (y)[c++]=(l)&255; (y)[c++]=((l)>>8)&255; \
                            memcpy((y)+c,(x),l); c+=l;
#define UNPACK_ITEM(x,y,c) { char *n=(char *)&(x); int i; \
        for(i=0;i<sizeof(x);i++) *(n++)=(y)[c++]; }
#define UNPACK_SLICE(x,l,y,c) (l)=(y)[c++]; (l)+=((y)[c++])<<8; \
                              memcpy((x),(y)+c,l); c+=l;

/*}}}*/
/*{{{  Structures defs and Statics*/
/* Used for temporary store of env stuff */
static char PseudoEnv[NoEnv][EnvSize];
static int CurrEnv=0;
/*}}}*/
/*{{{  Iserver socket tags (same as used by B300)*/
static enum Tags {
#if 0
	SpSocketA=70,
	SpSocketM=71,
#else
	SpSocketA=200,
	SpSocketM=201,
#endif
	SpAccept=5,
	SpBind=3,
	SpSockclose=17,
	SpConnect=2,
	SpGethostbyname=19,
	SpGethostbyaddr=20,
	SpSethostent=21,
	SpEndhostent=22,
	SpGethostname=33,
	SpSethostname=34,
	SpGetnetbyname=24,
	SpGetnetbyaddr=25,
	SpSetnetent=26,
	SpEndnetent=27,
	SpGetpeername=35,
	SpGetprotoent=28,
	SpGetprotobyname=29,
	SpGetprotobynum=30,
	SpSetprotoent=31,
	SpEndprotoent=32,
	SpGetsockname=36,
	SpGetsockopt=14,
	SpSetsockopt=15,
	SpIoctl=13,
	SpListen=4,
	SpReadsegstart=50,
	SpReadsegmore=51,
	SpRecvitsegstart=53,
	SpRecvitsegmore=54,
	SpSelect=12,
	SpSenditsegstart=56,
	SpSenditsegmore=57,
	SpShutdown=16,
	SpSocket=1,
	SpWritesegstart=59,
	SpWritesegmore=60,

	/*	The following tags are not specified by INMOS
*/
	SpInetAddr=43,
	SpGettimeofday=49,
	SpGetpwnam=45,
	SpGetpwuid=46,
	SpSockread=10,
	SpSockreadv=48,
	SpRecvit=7,
	SpSendit=9,
	SpSockwrite=11,
	SpSockwritev=47,


	/*	Tags specified by Inmos but not used here
*/
	SpAbort=38,
	SpPurge=39,

	SpOpenDir=190,
	SpReadDir=191,
	SpCloseDir=192,

	SpOpen2=173,
	SpSeek=17,
	SpCommand=40,
	SpTCL=193,
	SpError=20,

	/** TEMP FIX FOR T9000 **/
	SpGetuid=178,
	SpUid=1,
	SpEuid=2,
	SpGid=3,
	SpEgid=4,
	SpExit=35,
	SpStat=176

};
/*}}}*/
/*{{{  global vars*/
#ifndef XCC
char *sys_errlist[]={
	"Transputer X library : error messages absent",
	NULL };
int   sys_nerr=1;
#endif
static unsigned char Send[SpBuffSize],Recv[SpBuffSize];
/*}}}*/

/*{{{  networking ordering, byte manip functions*/
/*{{{  htons()*/
short htons(short in)
{ 
	short out;
	char *inptr=(char *)&in, *outptr=(char *)&out;
	outptr[0]=inptr[1];
	outptr[1]=inptr[0];
	return(out);
}
/*}}}*/
/*{{{  ntohs()*/
short ntohs(short in)
{ 
	short out;
	char *inptr=(char *)&in, *outptr=(char *)&out;
	outptr[0]=inptr[1];
	outptr[1]=inptr[0];
	return(out);
}
/*}}}*/
/*{{{  htonl()*/
long htonl(long in)
{ 
	long out;
	char *inptr=(char *)&in, *outptr=(char *)&out;
	outptr[0]=inptr[3];
	outptr[1]=inptr[2];
	outptr[2]=inptr[1];
	outptr[3]=inptr[0];
	return(out);
}
/*}}}*/
/*{{{  ntohl()*/
long ntohl(long in)
{ 
	long out;
	char *inptr=(char *)&in, *outptr=(char *)&out;
	outptr[0]=inptr[3];
	outptr[1]=inptr[2];
	outptr[2]=inptr[1];
	outptr[3]=inptr[0];
	return(out);
}
/*}}}*/

/*{{{  bcopy()*/
#ifndef FORTRAN
void bcopy(char *b1, char *b2, int length)
{
	memmove(b2,b1,(size_t)length);
}
#endif
/*}}}*/
/*{{{  bzero()*/
void bzero(char *b, int length)
{
	memset(b,0,(size_t)length);
}
/*}}}*/
/*}}}*/
/*{{{  random useful functions*/
/*{{{  int cfree(void *ptr)*/
int cfree(void *ptr)
{
	free(ptr);
	return(0);
}
/*}}}*/
/*}}}*/
/*{{{  accept()*/
int accept(int SockId, struct sockaddr *Addr, int *AddrLen)
{
	int rc, c, Len, OutSigio, NewErrno, NewSockId;
	short Family;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpAccept;
	PACK_ITEM(SockId,Send,c)
	Len=*AddrLen;
	PACK_ITEM(Len,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc==0)
	{ 
		char *AddrPtr=(char *)Addr;
		UNPACK_ITEM(NewSockId,Recv,c)
		rc=NewSockId;
		UNPACK_ITEM(Family,Recv,c)
		UNPACK_SLICE(AddrPtr,Len,Recv,c)
		Addr->sa_family=Family;
		*AddrLen=Len;
	}
	else { 
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	/*}}}*/
	return(rc);
}
/*}}}*/
/*{{{  bind()*/
int bind(int SockId, struct sockaddr *Addr, int AddrLen)
{
	int rc, c, Len, OutSigio, NewErrno;
	short Family;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpBind;
	PACK_ITEM(SockId,Send,c)
	Family=Addr->sa_family; 
	PACK_ITEM(Family,Send,c)
	PACK_SLICE(Addr,AddrLen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	return(rc);
}
/*}}}*/
/*{{{  open()*/
int open(char *path, int flags, int mode)
{
	int rc, c, Len, OutSigio, SockId, NewErrno;
	int oflags, spoflags, err;

#ifdef XCC
	oflags = 0; 
	spoflags = flags;
	if (spoflags & _O_RDONLY) oflags |= SP_O_RDONLY;
	if (spoflags & _O_WRONLY) oflags |= SP_O_WRONLY;
	if (spoflags & _O_RDWR  ) oflags |= SP_O_RDWR  ;
#ifdef  _O_NDELAY
	if (spoflags & _O_NDELAY) oflags |= SP_O_NDELAY;
#endif
	if (spoflags & _O_APPEND) oflags |= SP_O_APPEND;
#ifdef  _O_SYNC
	if (spoflags & _O_SYNC  ) oflags |= SP_O_SYNC  ;
#endif
	if (spoflags & _O_CREAT ) oflags |= SP_O_CREAT ;
	if (spoflags & _O_TRUNC ) oflags |= SP_O_TRUNC ;
	if (spoflags & _O_EXCL ) oflags |= SP_O_EXCL ;
#ifdef  _O_BINARY
	if (spoflags & _O_BINARY) oflags |= SP_O_BINARY;
#endif
#else
	oflags = flags;
#endif
	c=0;
	Send[c++]=SpOpen2;
	PACK_SLICE(path,strlen(path),Send,c)
	PACK_ITEM(oflags,Send,c)
	PACK_ITEM(mode,Send,c)
	c=(c+1) & (~1);
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(SockId,Recv,c)
	if ( rc == 0 )
		return(SockId);
	else {
		SockId = -1;
		c=0;
		Send[c++]=SpError;
		PACK_ITEM(SockId,Send,c)
		c=(c+1) & (~1);
		if ((Len=server_transaction(Send,c,Recv))<0)
		{ 
			errno=EIO; 
			return(-1); 
		}
		c=0;
		rc=Recv[c++];
		UNPACK_ITEM(err,Recv,c)
		errno = err;
		return(-1);
	}
}
/*}}}*/
/*{{{  creat()*/
int creat(char *path, int mode)
{ 
	int rc;
	rc=open(path,_O_WRONLY | _O_CREAT | _O_TRUNC,mode);
	return(rc);
}
/*}}}*/
/*{{{  close()*/
int close(int SockId)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpSockclose;
	PACK_ITEM(SockId,Send,c)
	if (c<8) c=8;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	rc=(rc==0) ? 0 : (-1);
	return(rc);
}
/*}}}*/
/*{{{  inet_addr()*/
unsigned long inet_addr(char *cp)
{
	int c, Len, OutSigio, NewErrno;
	unsigned long AddrVal, rc;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpInetAddr;
	PACK_SLICE(cp,strlen(cp),Send,c)
	if (c<8) c=8;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(AddrVal,Recv,c)
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	rc=(rc==0) ? AddrVal : (-1);
	return(rc);
}
/*}}}*/
/*{{{  sleep()*/
int sleep(unsigned DozePeriod)
{
	ProcWait(15625*DozePeriod);
	return(0);
}
/*}}}*/
/*{{{  ioctl()*/
int ioctl(int fd, int request, int *arg)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpIoctl;
	PACK_ITEM(fd,Send,c)
	PACK_ITEM(request,Send,c)
	if (arg!=NULL) { 
		int tmp=arg[0];
		Send[c++]=(char)1;
		PACK_ITEM(tmp,Send,c)
	}
	else Send[c++]=(char)0;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (arg!=NULL) { 
		int tmp;
		UNPACK_ITEM(tmp,Recv,c)
		arg[0]=tmp;
	}
	/*}}}*/
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1; 
	}
	return(rc);
}
/*}}}*/
/*{{{  getpwnam()*/
struct passwd *getpwnam(char *name)
{ 
	char *Buff;
	int i, rc, c, Len, OutSigio, NewErrno, Maxlen, Slen;
	struct passwd *Build=(struct passwd *)PseudoEnv[CurrEnv];

	if ((++CurrEnv)>=NoEnv) CurrEnv=0;   /* Wrap environment space */
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGetpwnam;
	Len=(name!=NULL) ? strlen(name)+1 : 0;
	PACK_SLICE(name,Len,Send,c)
	Maxlen=EnvSize-16;
	PACK_ITEM(Maxlen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(NULL); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		return(NULL); 
	}
	Buff=(char *)Build+sizeof(struct passwd);
	/*{{{  get pw_name*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_name=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get pw_passwd*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_passwd=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	UNPACK_ITEM(i,Recv,c); 
	Build->pw_uid=i;
	UNPACK_ITEM(i,Recv,c); 
	Build->pw_gid=i;
	/*{{{  get pw_age*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_age=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get pw_comment*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_comment=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get pw_gecos*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_gecos=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get pw_dir*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_dir=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get pw_shell*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_shell=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*}}}*/

	return(Build);
}
/*}}}*/
/*{{{  getpwuid()*/
struct passwd *getpwuid(int uid)
{ 
	char *Buff;
	int i, rc, c, Len, OutSigio, NewErrno, Maxlen, Slen;
	struct passwd *Build=(struct passwd *)PseudoEnv[CurrEnv];

	if ((++CurrEnv)>=NoEnv) CurrEnv=0;   /* Wrap environment space */
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGetpwuid;
	PACK_ITEM(uid,Send,c)
	Maxlen=EnvSize-16;
	PACK_ITEM(Maxlen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(NULL); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		return(NULL); 
	}
	Buff=(char *)Build+sizeof(struct passwd);
	/*{{{  get pw_name*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_name=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get pw_passwd*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_passwd=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	UNPACK_ITEM(i,Recv,c); 
	Build->pw_uid=i;
	UNPACK_ITEM(i,Recv,c); 
	Build->pw_gid=i;
	/*{{{  get pw_age*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_age=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get pw_comment*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_comment=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get pw_gecos*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_gecos=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get pw_dir*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_dir=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get pw_shell*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->pw_shell=(Slen==0) ? NULL : Buff;
	Buff+=Slen;
	/*}}}*/
	/*}}}*/

	return(Build);
}
/*}}}*/
/*{{{  gettimeofday()*/
int gettimeofday(struct timeval *tp, struct timezone *tzp)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGettimeofday;
	PACK_ITEM((tp->tv_sec),Send,c)
	PACK_ITEM((tp->tv_usec),Send,c)
	PACK_ITEM((tzp->tz_minuteswest),Send,c)
	PACK_ITEM((tzp->tz_dsttime),Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(0); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		return(-1); 
	}
	UNPACK_ITEM((tp->tv_sec),Recv,c)
	UNPACK_ITEM((tp->tv_usec),Recv,c)
	UNPACK_ITEM((tzp->tz_minuteswest),Recv,c)
	UNPACK_ITEM((tzp->tz_dsttime),Recv,c)
	/*}}}*/

	return(0);
}
/*}}}*/
/*{{{  connect()*/
int connect(int SockId, struct sockaddr *Addr, int AddrLen)
{
	int rc, c, Len, OutSigio, NewErrno;
	short Family;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpConnect;
	PACK_ITEM(SockId,Send,c)
	Family=Addr->sa_family; 
	PACK_ITEM(Family,Send,c)
	PACK_SLICE(Addr,AddrLen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}

	return(rc);
}
/*}}}*/
/*{{{  gethostbyname()*/
struct hostent *gethostbyname(char *Name)
{ 
	char *Buff;
	int rc, c, Len, OutSigio, NewErrno, Maxlen, Nalias, Naddr, i, Slen;
	struct hostent *Build=(struct hostent *)PseudoEnv[CurrEnv];

	if ((++CurrEnv)>=NoEnv) CurrEnv=0;   /* Wrap environment space */
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGethostbyname;
	Maxlen=SpBuffSize-20;
	for(Len=0;(Len<Maxlen) && (Name[Len]);Len++) ;
	Len++;
	PACK_SLICE(Name,Len,Send,c)
	Maxlen=EnvSize-16;
	PACK_ITEM(Maxlen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(NULL); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		return(NULL); 
	}
	Buff=(char *)Build+sizeof(struct hostent);
	Build->h_aliases=(char **)Buff;
	UNPACK_ITEM(Nalias,Recv,c)
	Build->h_addr_list=Build->h_aliases+(Nalias+1);
	UNPACK_ITEM(Naddr,Recv,c)
	Buff+=(Nalias+Naddr+2)*sizeof(char *);
	/*{{{  get h_name*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->h_name=Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get h_aliases*/
	for(i=0;i<Nalias;i++) {
		UNPACK_SLICE(Buff,Slen,Recv,c)
		Build->h_aliases[i]=Buff;
		Buff+=Slen;
	}
	Build->h_aliases[Nalias]=NULL;
	/*}}}*/
	UNPACK_ITEM(i,Recv,c); 
	Build->h_addrtype=i;
	UNPACK_ITEM(i,Recv,c); 
	Build->h_length=i;
	/*{{{  get h_addr_list*/
	for(i=0;i<Naddr;i++) {
		UNPACK_SLICE(Buff,Slen,Recv,c)
		Build->h_addr_list[i]=Buff;
		Buff+=Slen;
	}
	Build->h_addr_list[Naddr]=NULL;
	/*}}}*/
	/*}}}*/
	return(Build);
}
/*}}}*/
/*{{{  gethostbyaddr()*/
struct hostent *gethostbyaddr(char *Addr,int AddrLen, int Type)
{ 
	char *Buff;
	int rc, c, Len, OutSigio, NewErrno, Maxlen, Nalias, Naddr, i, Slen;
	struct hostent *Build=(struct hostent *)PseudoEnv[CurrEnv];

	if ((++CurrEnv)>=NoEnv) CurrEnv=0;   /* Wrap environment space */
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGethostbyaddr;
	PACK_SLICE(Addr,AddrLen,Send,c)
	PACK_ITEM(Type,Send,c)
	Maxlen=EnvSize-16;
	PACK_ITEM(Maxlen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(NULL); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		return(NULL); 
	}
	Buff=(char *)Build+sizeof(struct hostent);
	Build->h_aliases=(char **)Buff;
	UNPACK_ITEM(Nalias,Recv,c)
	Build->h_addr_list=Build->h_aliases+(Nalias+1);
	UNPACK_ITEM(Naddr,Recv,c)
	Buff+=(Nalias+Naddr+2)*sizeof(char *);
	/*{{{  get h_name*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->h_name=Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get h_aliases*/
	for(i=0;i<Nalias;i++) {
		UNPACK_SLICE(Buff,Slen,Recv,c)
		Build->h_aliases[i]=Buff;
		Buff+=Slen;
	}
	Build->h_aliases[Nalias]=NULL;
	/*}}}*/
	UNPACK_ITEM(i,Recv,c); 
	Build->h_addrtype=i;
	UNPACK_ITEM(i,Recv,c); 
	Build->h_length=i;
	/*{{{  get h_addr_list*/
	for(i=0;i<Naddr;i++) {
		UNPACK_SLICE(Buff,Slen,Recv,c)
		Build->h_addr_list[i]=Buff;
		Buff+=Slen;
	}
	Build->h_addr_list[Naddr]=NULL;
	/*}}}*/
	/*}}}*/
	return(Build);
}
/*}}}*/
/*{{{  sethostent()*/
void sethostent(int StayOpen)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpSethostent;
	PACK_ITEM(StayOpen,Send,c)
	if (c<8) c=8;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return; 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	return;
}
/*}}}*/
/*{{{  endhostent()*/
void endhostent(void)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpEndhostent;
	if (c<8) c=8;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return; 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	return;
}
/*}}}*/
/*{{{  gethostname()*/
int gethostname(char *Name, int NameLen)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGethostname;
	PACK_ITEM(NameLen,Send,c)
	if (c<8) c=8;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc==0)
	{
		UNPACK_SLICE(Name,Len,Recv,c)
	}
	else { 
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	/*}}}*/
	return(rc);
}
/*}}}*/
/*{{{  sethostname()*/
int sethostname(char *Name, int NameLen)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpSethostname;
	PACK_SLICE(Name,NameLen,Send,c)
	if (c<8) c=8;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (!rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	/*}}}*/
	return(rc);
}
/*}}}*/
/*{{{  getnetbyname()*/
struct netent *getnetbyname(char *Name)
{ 
	char *Buff;
	int rc, c, Len, OutSigio, NewErrno, Maxlen, Nalias, i, Slen;
	struct netent *Build=(struct netent *)PseudoEnv[CurrEnv];

	if ((++CurrEnv)>=NoEnv) CurrEnv=0;   /* Wrap environment space */
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGetnetbyname;
	Maxlen=SpBuffSize-20;
	for(Len=0;(Len<Maxlen) && (Name[Len]);Len++) ;
	Len++;
	PACK_SLICE(Name,Len,Send,c)
	Maxlen=EnvSize-16;
	PACK_ITEM(Maxlen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(NULL); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		return(NULL); 
	}
	Buff=(char *)Build+sizeof(struct netent);
	Build->n_aliases=(char **)Buff;
	UNPACK_ITEM(Nalias,Recv,c)
	Buff+=(Nalias+1)*sizeof(char *);
	/*{{{  get n_name*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->n_name=Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get n_aliases*/
	for(i=0;i<Nalias;i++) {
		UNPACK_SLICE(Buff,Slen,Recv,c)
		Build->n_aliases[i]=Buff;
		Buff+=Slen;
	}
	Build->n_aliases[Nalias]=NULL;
	/*}}}*/
	UNPACK_ITEM(i,Recv,c); 
	Build->n_addrtype=i;
	UNPACK_ITEM(i,Recv,c); 
	Build->n_net=i;
	/*}}}*/
	return(Build);
}
/*}}}*/
/*{{{  getnetbyaddr()*/
struct netent *getnetbyaddr(long Net, int Type)
{ 
	char *Buff;
	int rc, c, Len, OutSigio, NewErrno, Maxlen, Nalias, i, Slen;
	struct netent *Build=(struct netent *)PseudoEnv[CurrEnv];

	if ((++CurrEnv)>=NoEnv) CurrEnv=0;   /* Wrap environment space */
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGetnetbyaddr;
	PACK_ITEM(Net,Send,c)
	PACK_ITEM(Type,Send,c)
	Maxlen=EnvSize-16;
	PACK_ITEM(Maxlen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(NULL); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		return(NULL); 
	}
	Buff=(char *)Build+sizeof(struct netent);
	Build->n_aliases=(char **)Buff;
	UNPACK_ITEM(Nalias,Recv,c)
	Buff+=(Nalias+1)*sizeof(char *);
	/*{{{  get n_name*/
	UNPACK_SLICE(Buff,Slen,Recv,c)
	Build->n_name=Buff;
	Buff+=Slen;
	/*}}}*/
	/*{{{  get n_aliases*/
	for(i=0;i<Nalias;i++) {
		UNPACK_SLICE(Buff,Slen,Recv,c)
		Build->n_aliases[i]=Buff;
		Buff+=Slen;
	}
	Build->n_aliases[Nalias]=NULL;
	/*}}}*/
	UNPACK_ITEM(i,Recv,c); 
	Build->n_addrtype=i;
	UNPACK_ITEM(i,Recv,c); 
	Build->n_net=i;
	/*}}}*/
	return(Build);
}
/*}}}*/
/*{{{  setnetent()*/
void setnetent(int StayOpen)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpSetnetent;
	PACK_ITEM(StayOpen,Send,c)
	if (c<8) c=8;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return; 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	return;
}
/*}}}*/
/*{{{  endnetent()*/
void endnetent(void)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpEndnetent;
	if (c<8) c=8;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return; 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	return;
}
/*}}}*/
/*{{{  getpeername()*/
int getpeername(int SockId, struct sockaddr *Addr, int *AddrLen)
{
	int rc, c, Len, OutSigio, NewErrno;
	short Family;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGetpeername;
	PACK_ITEM(SockId,Send,c)
	Len=*AddrLen;
	PACK_ITEM(Len,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc==0)
	{ 
		char *AddrPtr=(char *)Addr;
		UNPACK_ITEM(Family,Recv,c)
		UNPACK_SLICE(AddrPtr,Len,Recv,c)
		Addr->sa_family=Family;
		*AddrLen=Len;
	}
	else { 
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	/*}}}*/
	return(rc);
}
/*}}}*/
/*{{{  setprotoent()*/
void setprotoent(int StayOpen)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpSetprotoent;
	PACK_ITEM(StayOpen,Send,c)
	if (c<8) c=8;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return; 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	return;
}
/*}}}*/
/*{{{  endprotoent()*/
void endprotoent(void)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpEndprotoent;
	if (c<8) c=8;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return; 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	return;
}
/*}}}*/
/*{{{  getsockname()*/
int getsockname(int SockId, struct sockaddr *Addr, int *AddrLen)
{
	int rc, c, Len, OutSigio, NewErrno;
	short Family;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGetsockname;
	PACK_ITEM(SockId,Send,c)
	Len=*AddrLen;
	PACK_ITEM(Len,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc==0)
	{ 
		char *AddrPtr=(char *)Addr;
		UNPACK_ITEM(Family,Recv,c)
		UNPACK_SLICE(AddrPtr,Len,Recv,c)
		Addr->sa_family=Family;
		*AddrLen=Len;
	}
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	return(rc);
}
/*}}}*/
/*{{{  setsockopt()*/
int setsockopt(int SockId, int Level, int OptName, char *OptVal, int OptLen)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpSetsockopt;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(Level,Send,c)
	PACK_ITEM(OptName,Send,c)
	if (OptVal==NULL) OptLen=0;
	PACK_SLICE(OptVal,OptLen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	return(rc);
}
/*}}}*/
/*{{{  listen()*/
int listen(int SockId, int Backlog)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpListen;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(Backlog,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	rc=(rc==0) ? SockId : (-1);
	return(rc);
}
/*}}}*/
/*{{{  recvfrom()*/
int recvfrom(int SockId, char *Data, int DataLen, int Flags,
struct sockaddr *From, int *FromLen)
{
	const int SpMaxData=SpBuffSize-16;
	int rc, c, Len, OutSigio, NewErrno, Nread, SegAddr, BytesToRead;
	short Family;

	/*{{{  Actually call the recv on the host, allocs mem on host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketM;
	Send[c++]=SpRecvit;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(DataLen,Send,c)
	PACK_ITEM(Flags,Send,c)
	if (From==NULL) *FromLen=0;
	Len=*FromLen;
	PACK_ITEM(Len,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(SegAddr,Recv,c)
	UNPACK_ITEM(Len,Recv,c)
	UNPACK_ITEM(BytesToRead,Recv,c)
	if (From!=NULL) {
		UNPACK_ITEM(Family,Recv,c)
		UNPACK_SLICE(From,Len,Recv,c)
		*FromLen=Len;
		From->sa_family=Family;
	}
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	if (rc) return(-1);
	/*}}}*/
	/*{{{  Read seg start - get first bit from host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpRecvitsegstart;
	PACK_ITEM(SegAddr,Send,c)
	PACK_ITEM(BytesToRead,Send,c)
	Len=(BytesToRead<=SpMaxData) ? BytesToRead : SpMaxData;
	PACK_ITEM(Len,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	if (rc==0)
	{
		UNPACK_SLICE(Data,Len,Recv,c)
		Nread=Len;
	}
	else return(-1);
	/*}}}*/
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		return(-1); 
	}
	/*}}}*/
	while (Nread<BytesToRead)
	{
		/*{{{  Read seg more - get next bit to host*/
		/*{{{  Fill send buffer*/
		c=0;
		Send[c++]=SpSocketA;
		Send[c++]=SpRecvitsegmore;
		PACK_ITEM(SegAddr,Send,c)
		Len=((BytesToRead-Nread)<=SpMaxData) ? (BytesToRead-Nread) : SpMaxData;
		PACK_ITEM(Len,Send,c)
		c=(c+1) & (~1);
		/*}}}*/
		if ((Len=server_transaction(Send,c,Recv))<0)
		{ 
			errno=EIO; 
			return(-1); 
		}
		/*{{{  Empty receive buffer*/
		c=0;
		rc=Recv[c++];
		UNPACK_ITEM(NewErrno,Recv,c)
		UNPACK_ITEM(OutSigio,Recv,c)
		if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
		if (rc && NewErrno) {
			PRINT_ERROR;
			errno=NewErrno;
		}
		if (rc==0)
		{
			UNPACK_SLICE(Data+Nread,Len,Recv,c)
			Nread+=Len;
		}
		else return(-1);
		/*}}}*/
		/*}}}*/
	}
	return(Nread);
}
/*}}}*/
/*{{{  read()*/
int read(int SockId, char *Data, int DataLen)
{
	const int SpMaxData=SpBuffSize-16;
	int rc, c, Len, OutSigio, NewErrno, Nread, SegAddr, BytesToRead;

	/*{{{  Actually call the read on the host, allocs mem on host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketM;
	Send[c++]=SpSockread;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(DataLen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(SegAddr,Recv,c)
	UNPACK_ITEM(Len,Recv,c)
	UNPACK_ITEM(BytesToRead,Recv,c)
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	if (rc) return(-1);
	/*}}}*/
	/*{{{  Read seg start - get first bit from host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpReadsegstart;
	PACK_ITEM(SegAddr,Send,c)
	PACK_ITEM(BytesToRead,Send,c)
	Len=(BytesToRead<=SpMaxData) ? BytesToRead : SpMaxData;
	PACK_ITEM(Len,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	if (rc==0)
	{
		UNPACK_SLICE(Data,Len,Recv,c)
		Nread=Len;
	}
	else return(-1);
	/*}}}*/
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		return(-1); 
	}
	/*}}}*/
	while (Nread<BytesToRead)
	{
		/*{{{  Read seg more - get next bit to host*/
		/*{{{  Fill send buffer*/
		c=0;
		Send[c++]=SpSocketA;
		Send[c++]=SpReadsegmore;
		PACK_ITEM(SegAddr,Send,c)
		Len=((BytesToRead-Nread)<=SpMaxData) ? (BytesToRead-Nread) : SpMaxData;
		PACK_ITEM(Len,Send,c)
		c=(c+1) & (~1);
		/*}}}*/
		if ((Len=server_transaction(Send,c,Recv))<0)
		{ 
			errno=EIO; 
			return(-1); 
		}
		/*{{{  Empty receive buffer*/
		c=0;
		rc=Recv[c++];
		UNPACK_ITEM(NewErrno,Recv,c)
		UNPACK_ITEM(OutSigio,Recv,c)
		if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
		if (rc && NewErrno) {
			PRINT_ERROR;
			errno=NewErrno;
		}
		if (rc==0)
		{
			UNPACK_SLICE(Data+Nread,Len,Recv,c)
			Nread+=Len;
		}
		else return(-1);
		/*}}}*/
		/*}}}*/
	}
	return(Nread);
}
/*}}}*/
/*{{{  readv()*/
int readv(int SockId, struct iovec *Iov, int IovCnt)
{ 
	char *Data;
	const int SpMaxData=SpBuffSize-16;
	int rc, c, Len, OutSigio, NewErrno, Nread, SegAddr, BytesToRead, i,
	TotData=0, CurrVec=0, CurrPos=0;

	if (IovCnt<=0) return(-1);
	for(i=0;i<IovCnt;i++) TotData+=Iov[i].iov_len;
	/*{{{  Actually call the read on the host, allocs mem on host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketM;
	Send[c++]=SpSockreadv;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(TotData,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(SegAddr,Recv,c)
	UNPACK_ITEM(Len,Recv,c)
	UNPACK_ITEM(BytesToRead,Recv,c)
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	if (rc) return(-1);
	/*}}}*/
	/*{{{  Read seg start - get first bit from host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpReadsegstart;
	PACK_ITEM(SegAddr,Send,c)
	PACK_ITEM(BytesToRead,Send,c)
	if ((Iov[CurrVec].iov_len-CurrPos)<=SpMaxData) {
		Len=Iov[CurrVec].iov_len-CurrPos;
		Data=Iov[CurrVec].iov_base+CurrPos;
		CurrVec++;
		CurrPos=0;
	}
	else {
		Len=SpMaxData;
		Data=Iov[CurrVec].iov_base+CurrPos;
		CurrPos+=SpMaxData;
	}
	PACK_ITEM(Len,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	if (rc==0)
	{
		UNPACK_SLICE(Data,Len,Recv,c)
		Nread=Len;
	}
	else return(-1);
	/*}}}*/
	if (rc) { 
		PRINT_ERROR;
		errno=NewErrno; 
		return(-1); 
	}
	/*}}}*/
	while (Nread<BytesToRead)
	{
		/*{{{  Read seg more - get next bit to host*/
		/*{{{  Fill send buffer*/
		c=0;
		Send[c++]=SpSocketA;
		Send[c++]=SpReadsegmore;
		PACK_ITEM(SegAddr,Send,c)
		if ((Iov[CurrVec].iov_len-CurrPos)<=SpMaxData) {
			Len=Iov[CurrVec].iov_len-CurrPos;
			Data=Iov[CurrVec].iov_base+CurrPos;
			CurrVec++;
			CurrPos=0;
		}
		else {
			Len=SpMaxData;
			Data=Iov[CurrVec].iov_base+CurrPos;
			CurrPos+=SpMaxData;
		}
		PACK_ITEM(Len,Send,c)
		c=(c+1) & (~1);
		/*}}}*/
		if ((Len=server_transaction(Send,c,Recv))<0)
		{ 
			errno=EIO; 
			return(-1); 
		}
		/*{{{  Empty receive buffer*/
		c=0;
		rc=Recv[c++];
		UNPACK_ITEM(NewErrno,Recv,c)
		UNPACK_ITEM(OutSigio,Recv,c)
		if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
		if (rc && NewErrno) {
			PRINT_ERROR;
			errno=NewErrno;
		}
		if (rc==0)
		{
			UNPACK_SLICE(Data,Len,Recv,c)
			Nread+=Len;
		}
		else return(-1);
		/*}}}*/
		/*}}}*/
	}
	return(Nread);
}
/*}}}*/
/*{{{  select()*/
int select(int Width, int *ReadFds, int *WriteFds, int *ExceptFds,
struct timeval *Timeout)
{
	int rc, c, Len, OutSigio, NewErrno, i, Words=((Width+31)/32), NoSelected;
	short ReadLen, WriteLen, ExceptLen;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpSelect;
	PACK_ITEM(Width,Send,c)
	ReadLen=(ReadFds!=NULL) ? Words : 0;
	PACK_ITEM(ReadLen,Send,c)
	for(i=0;i<ReadLen;i++) PACK_ITEM(ReadFds[i],Send,c)
		WriteLen=(WriteFds!=NULL) ? Words : 0;
	PACK_ITEM(WriteLen,Send,c)
	for(i=0;i<WriteLen;i++) PACK_ITEM(WriteFds[i],Send,c)
		ExceptLen=(ExceptFds!=NULL) ? Words : 0;
	PACK_ITEM(ExceptLen,Send,c)
	for(i=0;i<ExceptLen;i++) PACK_ITEM(ExceptFds[i],Send,c)
		if (Timeout!=NULL) {
			PACK_ITEM(Timeout->tv_sec,Send,c)
			PACK_ITEM(Timeout->tv_usec,Send,c)
		}
		else { 
			int tmp=-1;
			PACK_ITEM(tmp,Send,c)
			PACK_ITEM(tmp,Send,c)
		}
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	if (rc==0)
	{
		UNPACK_ITEM(NoSelected,Recv,c)
		UNPACK_ITEM(ReadLen,Recv,c)
		for(i=0;i<ReadLen;i++) UNPACK_ITEM(ReadFds[i],Recv,c)
			UNPACK_ITEM(WriteLen,Recv,c)
			for(i=0;i<WriteLen;i++) UNPACK_ITEM(WriteFds[i],Recv,c)
				UNPACK_ITEM(ExceptLen,Recv,c)
				for(i=0;i<ExceptLen;i++) UNPACK_ITEM(ExceptFds[i],Recv,c)
	}
	/*}}}*/
	if (rc || (NoSelected<0)) {
		PRINT_ERROR;
		errno=NewErrno; 
		NoSelected=-1;
	}
	return(NoSelected);
}
/*}}}*/
/*{{{  sendto()*/
int sendto(int SockId, char *Data, int DataLen,
int Flags, struct sockaddr *To, int ToLen)
{
	const int SpMaxData=SpBuffSize-16;
	int rc, c, Len, OutSigio, NewErrno, Written, SegAddr;
	short Family;

	/*{{{  Write seg start - alloc mem on host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpWritesegstart;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(DataLen,Send,c)
	Len=(DataLen<=SpMaxData) ? DataLen : SpMaxData;
	PACK_SLICE(Data,Len,Send,c)
	Written=Len;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(Len,Recv,c)
	UNPACK_ITEM(SegAddr,Recv,c)
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	if (rc) return(-1);
	/*}}}*/
	while (Written<DataLen)
	{
		/*{{{  Write seg more - send next bit to host*/
		/*{{{  Fill send buffer*/
		c=0;
		Send[c++]=SpSocketA;
		Send[c++]=SpWritesegmore;
		PACK_ITEM(SegAddr,Send,c)
		Len=((DataLen-Written)<=SpMaxData) ? (DataLen-Written) : SpMaxData;
		PACK_SLICE(Data+Written,Len,Send,c)
		Written+=Len;
		c=(c+1) & (~1);
		/*}}}*/
		if ((Len=server_transaction(Send,c,Recv))<0)
		{ 
			errno=EIO; 
			return(-1); 
		}
		/*{{{  Empty receive buffer*/
		c=0;
		rc=Recv[c++];
		UNPACK_ITEM(NewErrno,Recv,c)
		UNPACK_ITEM(OutSigio,Recv,c)
		if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
		UNPACK_ITEM(Len,Recv,c)
		/*}}}*/
		if (rc && NewErrno) {
			PRINT_ERROR;
			errno=NewErrno;
		}
		if (rc) return(-1);
		/*}}}*/
	}
	/*{{{  Actually call the send on the host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketM;
	Send[c++]=SpSendit;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(SegAddr,Send,c)
	Len=0;  /* Use this as offset - ignored by Sun iserver */
	PACK_ITEM(Len,Send,c)
	PACK_ITEM(DataLen,Send,c)
	PACK_ITEM(Flags,Send,c)
	if (To==NULL) ToLen=0;
	PACK_SLICE(To,ToLen,Send,c)
	if (ToLen>0) {
		Family=To->sa_family; 
		PACK_ITEM(Family,Send,c) 	}
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(Len,Recv,c)
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	rc=(rc==0) ? Len : (-1);
	return(rc);
	/*}}}*/
}
/*}}}*/
/*{{{  shutdown()*/
int shutdown(int SockId, int How)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpShutdown;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(How,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	return(rc);
}
/*}}}*/
/*{{{  socket()*/
int socket(int domain, int type, int protocol)
{
	int rc, c, Len, OutSigio, SockId, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpSocket;
	PACK_ITEM(domain,Send,c)
	PACK_ITEM(type,Send,c)
	PACK_ITEM(protocol,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(SockId,Recv,c)
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	rc=(rc==0) ? SockId : (-1);
	return(rc);
}
/*}}}*/
/*{{{  write()*/
int write(int SockId, char *Data, int DataLen)
{
	const int SpMaxData=SpBuffSize-16;
	int rc, c, Len, OutSigio, NewErrno, Written, SegAddr;

	/*{{{  Write seg start - alloc mem on host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpWritesegstart;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(DataLen,Send,c)
	Len=(DataLen<=SpMaxData) ? DataLen : SpMaxData;
	PACK_SLICE(Data,Len,Send,c)
	Written=Len;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(Len,Recv,c)
	UNPACK_ITEM(SegAddr,Recv,c)
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	if (rc) return(-1);
	/*}}}*/
	while (Written<DataLen)
	{
		/*{{{  Write seg more - send next bit to host*/
		/*{{{  Fill send buffer*/
		c=0;
		Send[c++]=SpSocketA;
		Send[c++]=SpWritesegmore;
		PACK_ITEM(SegAddr,Send,c)
		Len=((DataLen-Written)<=SpMaxData) ? (DataLen-Written) : SpMaxData;
		PACK_SLICE(Data+Written,Len,Send,c)
		Written+=Len;
		c=(c+1) & (~1);
		/*}}}*/
		if ((Len=server_transaction(Send,c,Recv))<0)
		{ 
			errno=EIO; 
			return(-1); 
		}
		/*{{{  Empty receive buffer*/
		c=0;
		rc=Recv[c++];
		UNPACK_ITEM(NewErrno,Recv,c)
		UNPACK_ITEM(OutSigio,Recv,c)
		if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
		UNPACK_ITEM(Len,Recv,c)
		/*}}}*/
		if (rc && NewErrno) {
			PRINT_ERROR;
			errno=NewErrno;
		}
		if (rc) return(-1);
		/*}}}*/
	}
	/*{{{  Actually call the write on the host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketM;
	Send[c++]=SpSockwrite;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(SegAddr,Send,c)
	Len=0;  /* Use this as offset - ignored by Sun iserver */
	PACK_ITEM(Len,Send,c)
	PACK_ITEM(DataLen,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(Len,Recv,c)
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	rc=(rc==0) ? Len : (-1);
	return(rc);
	/*}}}*/
}
/*}}}*/
/*{{{  writev()*/
int writev(int SockId, struct iovec *Iov, int IovCnt)
{ 
	char *Data;
	const int SpMaxData=SpBuffSize-16;
	int rc, c, Len, OutSigio, NewErrno, Written, SegAddr, i, TotData=0,
	CurrVec=0, CurrPos=0;


	if (IovCnt<=0) return(-1);
	for(i=0;i<IovCnt;i++) TotData+=Iov[i].iov_len;
	/*{{{  Write seg start - alloc mem on host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpWritesegstart;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(TotData,Send,c)
	if ((Iov[CurrVec].iov_len-CurrPos)<=SpMaxData) {
		Len=Iov[CurrVec].iov_len-CurrPos;
		Data=Iov[CurrVec].iov_base+CurrPos;
		CurrVec++;
		CurrPos=0;
	}
	else {
		Len=SpMaxData;
		Data=Iov[CurrVec].iov_base+CurrPos;
		CurrPos+=SpMaxData;
	}
	PACK_SLICE(Data,Len,Send,c)
	Written=Len;
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(Len,Recv,c)
	UNPACK_ITEM(SegAddr,Recv,c)
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	if (rc) return(-1);
	/*}}}*/
	while (Written<TotData)
	{
		/*{{{  Write seg more - send next bit to host*/
		/*{{{  Fill send buffer*/
		c=0;
		Send[c++]=SpSocketA;
		Send[c++]=SpWritesegmore;
		PACK_ITEM(SegAddr,Send,c)
		if ((Iov[CurrVec].iov_len-CurrPos)<=SpMaxData) {
			Len=Iov[CurrVec].iov_len-CurrPos;
			Data=Iov[CurrVec].iov_base+CurrPos;
			CurrVec++;
			CurrPos=0;
		}
		else {
			Len=SpMaxData;
			Data=Iov[CurrVec].iov_base+CurrPos;
			CurrPos+=SpMaxData;
		}
		PACK_SLICE(Data,Len,Send,c)
		Written+=Len;
		c=(c+1) & (~1);
		/*}}}*/
		if ((Len=server_transaction(Send,c,Recv))<0)
		{ 
			errno=EIO; 
			return(-1); 
		}
		/*{{{  Empty receive buffer*/
		c=0;
		rc=Recv[c++];
		UNPACK_ITEM(NewErrno,Recv,c)
		UNPACK_ITEM(OutSigio,Recv,c)
		if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
		UNPACK_ITEM(Len,Recv,c)
		/*}}}*/
		if (rc && NewErrno) {
			PRINT_ERROR;
			errno=NewErrno;
		}
		if (rc) return(-1);
		/*}}}*/
	}
	/*{{{  Actually call the writev on the host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketM;
	Send[c++]=SpSockwritev;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(SegAddr,Send,c)
	Len=0;  /* Use this as offset - ignored by Sun iserver */
	PACK_ITEM(Len,Send,c)
	PACK_ITEM(TotData,Send,c)
	PACK_ITEM(IovCnt,Send,c)
	for(i=0;i<IovCnt;i++)
	{
		Len=Iov[i].iov_len;
		PACK_ITEM(Len,Send,c)
	}
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	UNPACK_ITEM(Len,Recv,c)
	/*}}}*/
	if (rc && NewErrno) {
		PRINT_ERROR;
		errno=NewErrno;
	}
	rc=(rc==0) ? Len : (-1);
	return(rc);
	/*}}}*/
}
/*}}}*/
#ifdef XCC
/*{{{  getsockopt()*/
int getsockopt(int SockId, int Level, int OptName, char *OptVal, int *OptLen)
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpSocketA;
	Send[c++]=SpGetsockopt;
	PACK_ITEM(SockId,Send,c)
	PACK_ITEM(Level,Send,c)
	PACK_ITEM(OptName,Send,c)
	Len = *OptLen;
	PACK_ITEM(Len,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(NewErrno,Recv,c)
	UNPACK_ITEM(OutSigio,Recv,c)
	if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
	/*}}}*/
	if (rc) {
		PRINT_ERROR;
		errno=NewErrno; 
		rc=-1;
	}
	else {
		UNPACK_SLICE(OptVal,Len,Recv,c)
		*OptLen = Len;
	}
	return(rc);
}
/*}}}*/
/*{{{  opendir()*/
DIR *opendir(char *path)
{
	int rc, c, Len, DirId, err;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpOpenDir;
	PACK_SLICE(path,strlen(path),Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(NULL); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(DirId,Recv,c)
	/*}}}*/
	if ( rc == 0 )
		return((DIR *)DirId);
	else {
		DirId = -1;
		c=0;
		Send[c++]=SpError;
		PACK_ITEM(DirId,Send,c)
		c=(c+1) & (~1);
		if ((Len=server_transaction(Send,c,Recv))<0)
		{ 
			errno=EIO; 
			return( NULL );
		}
		c=0;
		rc=Recv[c++];
		UNPACK_ITEM(err,Recv,c)
		errno = err;
		return( NULL );
	}
}
/*}}}*/
/*{{{  readdir()*/
struct dirent * readdir(int DirId)
{
	const int SpMaxData=SpBuffSize-16;
	int rc, c, Len, Nread, SegAddr, BytesToRead;
	static struct dirent dir_entry;

	/*{{{  Actually call the read on the host, allocs mem on host*/
	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpReadDir;
	PACK_ITEM(DirId,Send,c)
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(NULL); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	/*}}}*/
	if (rc == 0){
		unsigned long tempint;
		unsigned short tempsh;
		char *ptr;
		UNPACK_ITEM(tempint, Recv, c);
		dir_entry.d_off = tempint;
		UNPACK_ITEM(tempint, Recv, c);
		dir_entry.d_ino = tempint;
		UNPACK_ITEM(tempsh, Recv, c);
		dir_entry.d_reclen = tempint;
		UNPACK_ITEM(tempsh, Recv, c);
		dir_entry.d_namlen = tempint;
		tempint = dir_entry.d_namlen+1;
		ptr = (char *)&(dir_entry.d_name);
		UNPACK_SLICE( ptr,tempint, Recv, c );
		return( &dir_entry);


	}
	if (rc) return(NULL);
	/*}}}*/
}
/*}}}*/
/*{{{  closedir()*/
int closedir(unsigned int fd)
{
	int rc, c, Len;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpCloseDir;
	PACK_ITEM(fd,Send,c);
	c=(c+1) & (~1);
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	/*{{{  Empty receive buffer*/
	c=0;
	rc=Recv[c++];
	/*}}}*/
	rc=(rc==0) ? 0 : (-1);
	return(rc);
}
/*}}}*/

/*{{{  GetArgs()*/
#if 0
#define SWITCH_CHAR	'-'
void ParseCommandLine ( char *CommandLine )
{
	register unsigned char *s, *t, *str;
	int i;
	int Gobble=1;

	t = (char *)CommandLine;
	str = (unsigned char *)malloc( (strlen(CommandLine)+1)*sizeof(unsigned char) );
	s = str;
	strcpy( s, CommandLine );
	*t= '\0';
	while( (*s) && (*s != ' ') )
		++s;
	if ( *s == ' ' )
		++s;
	for(;;)
	{
		if ( *s == '\0' ){
			*t = '\0';
			free(str);
			return;
		}
		if (  (*s==SWITCH_CHAR) && ( (*(s+1)=='s') || (*(s+1)=='S') )  )
		{
			s += 2;
			switch ( *s )
			{
			case 'a' :
			case 'A' :  
				++s; 
				break;
			case 'b' :
			case 'B' :
				++s;
				while( *s == ' ' )
					++s;
				i = 0;
				while( (*s) && (*s != SWITCH_CHAR) && (*s != ' ') )
					*t++ = *s++;
				*t++ = ' ';
				break;
			case 'c' :
			case 'C' :
				++s;
				while( *s == ' ' )
					++s;
				while( (*s) && ( (*s!=SWITCH_CHAR) && (*s!=' ') ) )
					s++;
				break;
			case 'e' :
			case 'E' :  
				++s; 
				break;
			case 'i' :
			case 'I' :  
				++s; 
				break;
			case 'l' :
			case 'L' :  
				++s;
				while( *s == ' ' )
					++s;
				while( (*s) && ( (*s!=SWITCH_CHAR) && (*s!=' ') ) )
					s++;
				break;

			case 'p' :
			case 'P' :  
				++s;
				while (*s == ' ')
					++s;
				while( (*s) && (*s!=SWITCH_CHAR) )
					++s;
				break;
			case 'r' :
			case 'R' :  
				++s; 
				break;
			case 's' :
			case 'S' :  
				++s; 
				break;
			case 'z' :
			case 'Z' :  
				++s; 
				break;
			default  :
				*t++ = SWITCH_CHAR;
				*t++ = *(s-1);
				Gobble = 0;
				break;
			}
			if ( (Gobble) && (*s == ' ') )
				++s;
		}
		else
		{
			if (  (*s==SWITCH_CHAR) && ( (*(s+1)=='p') || (*(s+1)=='P') )  )
			{
				int i=0;
				s+=2; /* skip -P or -p */
				while (*s==' ') s++;
				while ( (*s) && ( ( *s!=SWITCH_CHAR)  && (*s!=' ') ) )
					s++;

			}
			else
			{
				*t++ = *s++;
			}
		}
	}
	free( str );
}
int GetArgs( char *str )
{
	int rc, c, Len, OutSigio, NewErrno;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpCommand;
	Send[c++]=1;
	c=(c+1) & (~1);
	c = (c < 8)?8:0;
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	c=0;
	rc=Recv[c++];
	/*
  if ( rc != 0 )return( 0 );
  */
	UNPACK_SLICE( str,Len, Recv, c );
	str[Len] = '\0';
	ParseCommandLine(str);
	return(Len);
}
#endif
/*}}}*/


/*{{{  server_transaction*/

#ifndef T9000
int server_transaction( unsigned char *Send, int len, unsigned char *Recv )
{
	unsigned short size = len;

	__semp( &_Node.serversem );
	if ( size < 8 ) size = 8;
	ChanOut( _Node.toserver, &size, 2 );
	ChanOut( _Node.toserver, Send, size );
	ChanIn( _Node.fromserver, &size, 2 );
	if ( size > 0 )
		ChanIn( _Node.fromserver, Recv, size );
	__semv( &_Node.serversem );
	return size;
}

#else
#if 0

void VChanIn( Channel *InChan, char *Buffer, int BufSize )
{
	__asm( "ldl 4" );	/* SYS_fnent_argptr + tempWorkSpaceSize */
	__asm( "ldnl 1" );	/* Buffer */
	__asm( "ldl 4" );	/* SYS_fnent_argptr + tempWorkSpaceSize */
	__asm( "ldnl 0" );	/* InChan */
	__asm( "ldl 4" );	/* SYS_fnent_argptr + tempWorkSpaceSize */
	__asm( "ldnl 2" );	/* BufSize */
	__asm( "vin" );
}

int server_transaction( unsigned char *Send, int len, unsigned char *Recv )
{
	unsigned short size = len;

	__semp( &_Node.serversem );
	ChanOut( _Node.toserver, Send, size );
	VChanIn( _Node.fromserver, Recv, SpBuffSize );
	__semv( &_Node.serversem );
	return size;
}
#endif


/** TEMP FIX **/
getuid()
{
	int rc, c, Len;
	int uid;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpGetuid;
	Send[c++]=SpUid;
	c=(c+1) & (~1);
	c = (c < 8)?8:0;
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(uid,Recv,c)
	return(uid);
}

geteuid()
{
	int rc, c, Len;
	int uid;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpGetuid;
	Send[c++]=SpEuid;
	c=(c+1) & (~1);
	c = (c < 8)?8:0;
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(uid,Recv,c)
	return(uid);
}

getgid()
{
	int rc, c, Len;
	int uid;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpGetuid;
	Send[c++]=SpGid;
	c=(c+1) & (~1);
	c = (c < 8)?8:0;
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(uid,Recv,c)
	return(uid);
}

getegid()
{
	int rc, c, Len;
	int uid;

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpGetuid;
	Send[c++]=SpEgid;
	c=(c+1) & (~1);
	c = (c < 8)?8:0;
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	c=0;
	rc=Recv[c++];
	UNPACK_ITEM(uid,Recv,c)
	return(uid);
}

#if 0
__Exit(int retval)
{
	int rc, c, Len;
	unsigned char Send[20], Recv[20]; /* TEMP FIXME */

	/*{{{  Fill send buffer*/
	c=0;
	Send[c++]=SpExit;
	if ( retval == 0 )retval = 999999999;
	else retval = -999999999;
	PACK_ITEM(retval,Send,c);
	c=(c+1) & (~1);
	c = (c < 8)?8:0;
	/*}}}*/
	if ((Len=server_transaction(Send,c,Recv))<0)
	{ 
		errno=EIO; 
		return(-1); 
	}
	return(0);
}
#endif


/*{{{  stat()*/
#define  SP_S_IFMT       0170000
#define SP_S_IFIFO      0040000
#define SP_S_IFCHR      0020000
#define SP_S_IFDIR      0060000
#define SP_S_IFBLK      0100000
#define SP_S_IFREG      0010000
#define SP_S_ISUID      04000
#define SP_S_ISGID      02000
#define SP_S_ISVTX      01000           /* SP_STAT mode */
#define SP_S_ENFMT      SP_S_ISGID
#define SP_S_IRUSR      00400
#define SP_S_IWUSR      00200
#define SP_S_IXUSR      00100
#define SP_S_IRGRP      00040
#define SP_S_IWGRP      00020
#define SP_S_IXGRP      00010
#define SP_S_IROTH      00004
#define SP_S_IWOTH      00002
#define SP_S_IXOTH      00001

#define _IFMT           0170000 /* type of file */
#define         _IFDIR  0040000 /* directory */
#define         _IFCHR  0020000 /* character special */
#define         _IFBLK  0060000 /* block special */
#define         _IFREG  0100000 /* regular */
#define         _IFLNK  0120000 /* symbolic link */
#define         _IFSOCK 0140000 /* socket */
#define         _IFIFO  0010000 /* fifo */

#define _S_ISUID         0004000 /* set user id on execution */
#define _S_ISGID         0002000 /* set group id on execution */
#ifndef _POSIX_SOURCE
#define _S_ISVTX         0001000 /* save swapped text even after use */
#define _S_IREAD         0000400 /* read permission, owner */
#define _S_IWRITE        0000200 /* write permission, owner */
#define _S_IEXEC         0000100 /* execute/search permission, owner */

#define _S_ENFMT         0002000 /* enforcement-mode locking */

#define _S_IFMT          _IFMT
#define _S_IFDIR         _IFDIR
#define _S_IFCHR         _IFCHR
#define _S_IFBLK         _IFBLK
#define _S_IFREG         _IFREG
#define _S_IFLNK         _IFLNK
#define _S_IFSOCK        _IFSOCK
#define _S_IFIFO         _IFIFO
#endif  !_POSIX_SOURCE

#define _S_IRWXU         0000700 /* rwx, owner */
#define         _S_IRUSR 0000400 /* read permission, owner */
#define         _S_IWUSR 0000200 /* write permission, owner */
#define         _S_IXUSR 0000100 /* execute/search permission, owner */
#define _S_IRWXG         0000070 /* rwx, group */
#define         _S_IRGRP 0000040 /* read permission, group */
#define         _S_IWGRP 0000020 /* write permission, grougroup */
#define         _S_IXGRP 0000010 /* execute/search permission, group */
#define _S_IRWXO         0000007 /* rwx, other */
#define         _S_IROTH 0000004 /* read permission, other */
#define         _S_IWOTH 0000002 /* write permission, other */
#define         _S_IXOTH 0000001 /* execute/search permission, other */



int stat(char *path, struct stat *buf)
{
  int rc, c, Len;
  int fd=-1;
  int mode;

  /*{{{  Fill send buffer*/
  c=0;
  Send[c++]=SpStat;
  PACK_SLICE(path,strlen(path),Send,c)
  PACK_ITEM(fd,Send,c);
  if (c<8) c=8;
  c=(c+1) & (~1);
  /*}}}*/
  if ((Len=server_transaction(Send,c,Recv))<0)
     { errno=EIO; return(-1); }
  /*{{{  Empty receive buffer*/
  c=0;
  rc=Recv[c++];
  if
    (rc==0)
    {
#ifndef XCC
      UNPACK_ITEM(buf->st_dev,Recv,c)
      UNPACK_ITEM(buf->st_ino,Recv,c)
      UNPACK_ITEM(buf->st_mode,Recv,c)
      UNPACK_ITEM(buf->st_nlink,Recv,c)
      UNPACK_ITEM(buf->st_uid,Recv,c)
      UNPACK_ITEM(buf->st_gid,Recv,c)
      UNPACK_ITEM(buf->st_rdev,Recv,c)
      UNPACK_ITEM(buf->st_size,Recv,c)
      UNPACK_ITEM(buf->st_atime,Recv,c)
      UNPACK_ITEM(buf->st_mtime,Recv,c)
      UNPACK_ITEM(buf->st_ctime,Recv,c)
      UNPACK_ITEM(buf->st_blksize,Recv,c)
      UNPACK_ITEM(buf->st_blocks,Recv,c)
#else
      unsigned long temp;
      UNPACK_ITEM(buf->st_mode,Recv,c)
      UNPACK_ITEM(temp,Recv,c)
      buf->st_ino = temp;
      UNPACK_ITEM(buf->st_dev,Recv,c)
      UNPACK_ITEM(buf->st_rdev,Recv,c)
      UNPACK_ITEM(buf->st_nlink,Recv,c)
      UNPACK_ITEM(buf->st_uid,Recv,c)
      UNPACK_ITEM(buf->st_gid,Recv,c)
      UNPACK_ITEM(buf->st_size,Recv,c)
      UNPACK_ITEM(buf->st_atime,Recv,c)
      UNPACK_ITEM(buf->st_mtime,Recv,c)
      UNPACK_ITEM(buf->st_ctime,Recv,c)
  

    mode = 0;
    switch (buf->st_mode & SP_S_IFMT) {
    case SP_S_IFIFO:   mode = _S_IFIFO;  break;
    case SP_S_IFCHR:   mode = _S_IFCHR;  break;
    case SP_S_IFDIR:   mode = _S_IFDIR;  break;
    case SP_S_IFBLK:   mode = _S_IFBLK;  break;
    case SP_S_IFREG:   mode = _S_IFREG;  break;
    }
    if (buf->st_mode & SP_S_ISUID)    mode |= _S_ISUID;
    if (buf->st_mode & SP_S_ISGID)    mode |= _S_ISGID;
    if (buf->st_mode & SP_S_ISVTX)    mode |= _S_ISVTX;
#ifdef SP_S_ENFMT
    if (buf->st_mode & SP_S_ENFMT)    mode |= _S_ENFMT;
#endif
    if (buf->st_mode & SP_S_IRUSR)    mode |= _S_IRUSR;
    if (buf->st_mode & SP_S_IWUSR)    mode |= _S_IWUSR;
    if (buf->st_mode & SP_S_IXUSR)    mode |= _S_IXUSR;
    if (buf->st_mode & SP_S_IRGRP)    mode |= _S_IRGRP;
    if (buf->st_mode & SP_S_IWGRP)    mode |= _S_IWGRP;
    if (buf->st_mode & SP_S_IXGRP)    mode |= _S_IXGRP;
    if (buf->st_mode & SP_S_IROTH)    mode |= _S_IROTH;
    if (buf->st_mode & SP_S_IWOTH)    mode |= _S_IWOTH;
    if (buf->st_mode & SP_S_IXOTH)    mode |= _S_IXOTH;
    buf->st_mode =  mode;

#endif
    }
  /*}}}*/
  if (rc!=0) rc=-1;
  return(rc);
}
/*}}}*/


#endif /* T9000 */

#ifdef FORTRAN
void raise( int signum )
{
}

#endif
#endif
/*}}}*/
