/*	Macros used by TCL and TCL-VCP interface routines
*/

/*{{{  REPORT TCL ERROR */
/*	Reports error and aborts the current function.
*/
#define REPORT_ERROR( interp, ErrorMessage, Parameter ) {		\
	Tcl_AppendResult( interp, ErrorMessage, Parameter, NULL );	\
	return( TCL_ERROR );						\
}

/*	Report error in SEND command format
*/
#define SEND_COMMAND_ERROR( interp )					\
	REPORT_ERROR( interp, "ERROR : send <address> [-data|-file] <source> [ -bind <command> ] [-timeout <time> <command>]\n", NULL )

/*	Report error in RECV command format
*/
#define RECV_COMMAND_ERROR( interp )					\
	REPORT_ERROR( interp, "ERROR : recv <address> [-data|-file] <destn> [ -bind <command> ] [-timeout <time> <command>]\n", NULL )
/*}}}*/

/*{{{  QUEUE_TCL_COMMAND*/
/*	Add a command to the TclCommandList. These commands will be
	executed when TCL is idle
*/
#if 0
#define QUEUE_TCL_COMMAND( CommandStr ) {				\
	extern Semaphore TclSemaphore;					\
	extern TCL_COMMAND *TclCommandList;				\
	extern int NumTclPending;					\
	TCL_COMMAND	*TclCommand;					\
	if ( *CommandStr != '\0' ){					\
		TclCommand = MEM_ALLOC( TCL_COMMAND );			\
		TclCommand->Command = CommandStr;			\
		SemWait( &TclSemaphore );				\
		EnQueue( TclCommandList, TclCommand );			\
		SemSignal( &TclSemaphore );				\
	}								\
	else {								\
		SemWait( &TclSemaphore );				\
		NumTclPending--;					\
		SemSignal( &TclSemaphore );				\
	}								\
}
#else
#define QUEUE_TCL_COMMAND( CommandStr ) {				\
	extern Semaphore TclSemaphore;					\
	extern TCL_COMMAND *TclCommandList;				\
	extern int NumTclPending;					\
	TCL_COMMAND	*TclCommand;					\
	TclCommand = MEM_ALLOC( TCL_COMMAND );				\
	TclCommand->Command = CommandStr;				\
	SemWait( &TclSemaphore );					\
	EnQueue( TclCommandList, TclCommand );				\
	SemSignal( &TclSemaphore );					\
}
#endif
/*}}}*/

/*{{{  GET_LINK_NUMBER*/
/*	Return the link number specified in the string 'Address'.
	'Address' is not modified. 
	Format of LinkNumber specification "[Ll][0-3]"
*/
#define	GET_LINK_NUMBER( Address, LinkNumber ) {			\
	if (( *Address == 'L' )||( *Address == 'l' ))			\
		LinkNumber = (*(Address+1) - '0' );			\
}
/*}}}*/

/*{{{  SKIP_LINK_NUMBER*/
/*	Skip the link number specified in 'Address'
	'Address' is made to point to the next token
	Format of LinkNumber specification "[Ll][0-3]"
*/
#define	SKIP_LINK_NUMBER( Address ) {					\
	if (( *Address == 'L' )||( *Address == 'l' )){			\
		while ( *Address != ADDRESS_SEPARATOR )			\
			Address++;					\
		Address++;						\
	}								\
}
/*}}}*/

/*{{{  GET_HEADER_VAL*/
/*	Get the size and value of a header
*/
#define GET_HEADER_VAL( Header, Val, size ) {				\
	Val = 0;							\
	size = (*Header == '0' ) ? 2:1;					\
	Val = StrToNum( Header, HEADER_RADIX );				\
	if (( Val == 0 )&&(*(Header+1) != '0' )) size = 1;		\
	size = ( Val > 255 ) ? 2:size;					\
}
/*}}}*/

/*{{{  SKIP_HEADER_VAL*/
/*	Skip the current header field
*/
#define SKIP_HEADER_VAL( Header ){					\
	while (( *Header != ADDRESS_SEPARATOR )&&(!END_OF_ARG(Header))) \
		Header++;						\
	if ( *Header == ADDRESS_SEPARATOR )Header++;			\
}
/*}}}*/

/*{{{  FORM_VIRTUAL_CHAN_ADDR*/
/*	Form a virtual channel address from a string of the form
	<Link><Address>
*/
#define	FORM_VIRTUAL_CHAN_ADDR( ReceiveAddress, VirtualChan ) {		\
	unsigned char *Address = ReceiveAddress;			\
	SKIP_LINK_NUMBER( Address );					\
	GET_HEADER_VAL( Address, VirtualChan.HeaderValue, VirtualChan.HeaderSize );	\
}
/*}}}*/

/*{{{  SKIP_ONE_HEADER*/
/*	Go to the next header 
	Address is modified.
*/
#define SKIP_ONE_HEADER( Address )					\
	while (!END_OF_ARG( Address ))					\
	       if ( *Address++ == ADDRESS_SEPARATOR )			\
			break;
/*}}}*/

/*{{{  FORM_HEADER*/
/*	Form a header from  a string of the form
	<Link>(<header 1>.)*<Header n>
*/
#define	FORM_HEADER( SendAddress, Header ) {				\
	unsigned char *Address = SendAddress;				\
	int val, size;							\
	int index = 0;							\
	unsigned char *HeaderPtr;					\
	SKIP_LINK_NUMBER(Address);					\
	Header.HeaderSize = 0;						\
	while (!END_OF_ARG( Address )) {				\
		GET_HEADER_VAL( Address, val, size );			\
		SKIP_HEADER_VAL( Address );				\
		Header.HeaderSize += size;				\
	}								\
	Header.HeaderPtr = (unsigned char *)vcp_malloc( Header.HeaderSize );\
	Address = SendAddress;						\
	SKIP_LINK_NUMBER( Address );					\
	HeaderPtr = Header.HeaderPtr;					\
	while (!END_OF_ARG( Address )){					\
		GET_HEADER_VAL( Address, val, size );			\
		SKIP_HEADER_VAL( Address );				\
		if ( size == 1 ){					\
			*HeaderPtr = (unsigned char)val;		\
			HeaderPtr++;					\
		}							\
		else {							\
			*(INT16 *)HeaderPtr = (INT16)val;		\
			HeaderPtr += 2;					\
		}							\
	}								\
}
/*}}}*/

/*{{{  WRITE_MESSAGE_TO_FILE*/
/*	Write a message to a file in ascii form
*/
#define	WRITE_MESSAGE_TO_FILE( buffer, bufsize, fp ){			\
	int i;								\
	for ( i = 0; i < bufsize; i++ ){				\
		if ( i % 10 )	fprintf( fp, "\n" );			\
		fprintf( fp, "0x%02x%c", buffer[i], DATA_SEPARATOR );	\
	}								\
}
/*}}}*/

/*{{{  MESSAGE_COPY*/
/*	Copy binary data from source into destination using the
	format <HexByte> <HexByte>...
*/
#define MESSAGE_COPY( destination, source, length ){			\
	int si, di;							\
	for ( si = 0, di = 0; si < length-1; si++, di+=LENGTH_PER_BYTE )\
		sprintf( &(destination[di]), "0x%02x%c", 		\
			(unsigned char)source[si], DATA_SEPARATOR );	\
	sprintf( &(destination[di]), "0x%02x",(unsigned char)source[si] );\
}
/*}}}*/

/*{{{  IS_HEXVAL*/
/*	Check if the given char is a valid hex digit
*/
#define IS_HEXVAL(c) ((( c >= '0') && ( c <= '9' ))||			\
		  (( c >= 'a') && ( c <= 'f' ))||			\
		  (( c >= 'A') && ( c <= 'F' )))
/*}}}*/

/*{{{  HEXVAL*/
/*	Return hex value of a digit
*/
#define HEXVAL(c) (( c >= '0') && ( c <= '9' ))?(c - '0'):		\
		  ((( c >= 'a') && ( c <= 'f' ))?(c +10 - 'a' ):	\
		   ( c + 10 - 'A' ))
/*}}}*/
		   
/*{{{  MAKE_MESSAGE*/
/*	From a string of form <HexByte> <HexByte> ..., form a binary
	message
*/
#if 0
#define MAKE_MESSAGE( destination, source, length, MessageLen ){	\
	int si=0, di=0;							\
	unsigned char temp;						\
	if (END_OF_ARG(source))						\
		MessageLen = 0;						\
	else {								\
		di = 1;							\
		while ( si < length )					\
			if ( source[si++] == DATA_SEPARATOR )di++;	\
		destination = vcp_malloc( di * sizeof(char));		\
		MessageLen = ( destination == NULL )?0:di;		\
		for ( si = 0, di = 0; di < MessageLen; di++, si++ ){	\
			temp = 0;					\
			while (( si < length )&&			\
			  ( source[si] != DATA_SEPARATOR )){		\
				temp = temp*16 + (HEXVAL(source[si]));	\
				si++;					\
			}						\
			destination[di] = temp;				\
		}							\
	}								\
}
#else
#define GET_ELEMENT_LEN( element, len ){				\
	int lastbyte = strlen(element)-1;				\
	if ( element[lastbyte] == LONG_WORD_SPEC )len = 4;		\
	else if ( element[lastbyte] == SHORT_WORD_SPEC )len = 2;	\
	else len = 1;							\
}

#define GET_MESSAGE_ELEMENT( message, element ){			\
	int i = 0, j = 0;						\
	while (!END_OF_ARG((&(message[i])))&&( message[i] != DATA_SEPARATOR ))\
		element[j++] = message[i++];				\
	element[j] = '\0';						\
}

#define GET_ELEMENT_VALUE( element, value ){				\
	if ( strncmp( element, "0x",2 ) == 0 )				\
		sscanf( &(element[2]), "%x", &value );			\
	else if ( element[0] == '0' )					\
		sscanf( element, "%o", &value );			\
	else								\
		sscanf( element, "%d", &value );			\
}

#define ADD_ELEMENT_TO_MESSAGE( msg, element, len ){			\
	unsigned char temp[MAX_MESSAGE_ELEMENT_LEN];			\
	int lastbyte = strlen(element)-1;				\
	unsigned long value;						\
	unsigned char *ptr;						\
	int i;								\
	strcpy( temp, element ); 					\
	if ( element[lastbyte] == LONG_WORD_SPEC ){			\
		len = 4;						\
		temp[lastbyte] = '\0';					\
	}								\
	else if ( element[lastbyte] == SHORT_WORD_SPEC ){		\
		len = 2;						\
		temp[lastbyte] = '\0';					\
	}								\
	else {								\
		len = 1;						\
	}								\
	GET_ELEMENT_VALUE( temp, value );				\
	ptr = (unsigned char *)&value;					\
	for ( i = 0; i < len; i++ )					\
		msg[i] = *ptr++;					\
}

#define MAKE_MESSAGE( destination, source, length, MessageLen ){	\
	int si = 0, di = 0;						\
	int len;							\
	unsigned char element[MAX_MESSAGE_ELEMENT_LEN];			\
	if (END_OF_ARG(source))						\
		MessageLen = 0;						\
	else {								\
		while ( si < length ){					\
			GET_MESSAGE_ELEMENT( (&(source[si])), element);	\
			si += strlen( element )+1;			\
			GET_ELEMENT_LEN( element, len );		\
			di += len;					\
		}							\
		destination = vcp_malloc( di * sizeof(char));		\
		if ( destination == NULL ){				\
			MessageLen = 0;					\
		}							\
		else {							\
			MessageLen = di;				\
			si = 0; di = 0;					\
			while ( si < length ){				\
				GET_MESSAGE_ELEMENT( (&(source[si])), element );\
				si += strlen( element )+1;		\
				ADD_ELEMENT_TO_MESSAGE( (&(destination[di])), element, len );\
				di += len;				\
			}						\
		}							\
	}								\
}
#endif
/*}}}*/

/*{{{  DECIMAL_DIGIT*/
/*	Check if c is a valid decimal digit
*/
#define DECIMAL_DIGIT( c ) (( c >='0' )&&( c <= '9' ))
/*}}}*/

/*{{{  HEX_DIGIT*/
/*	Check if c is a valid hexadecimal digit
*/
#define HEX_DIGIT( c )	((( c >= '0' )&&( c <= '9' ))||			\
			 (( c >= 'a' )&&( c <= 'f' ))||			\
			 (( c >= 'A' )&&( c <= 'F' )))
/*}}}*/
		
/*{{{  END_OR_ARG*/
/*	Check if string points to the end of an argument
*/
#define END_OF_ARG( string ) ((*(string) == '\0')||(*(string) == '\n'))
/*}}}*/

/*{{{  CHECK_ADDRESS_SEPARATOR*/
/*	Check if string points to a address separator
*/
#define	CHECK_ADDRESS_SEPARATOR( string, interp ){			\
	if (*string != ADDRESS_SEPARATOR )				\
		REPORT_ERROR( interp,  "Expected address separator", 	\
			NULL );						\
}
/*}}}*/

/*{{{  CHECK_ONE_HEADER*/
/*	Check if string points to a valid header. Return the length of
 	the header in HeaderLen
*/
#define	CHECK_ONE_HEADER( string, HeaderLen, interp ){			\
	unsigned char *HeaderStr = string;				\
	int val = 0;							\
	HeaderLen = 0;							\
	if ( !DECIMAL_DIGIT( *HeaderStr ))				\
		REPORT_ERROR( interp,  "Expected decimal value in header", \
			NULL );						\
	while ( DECIMAL_DIGIT( *HeaderStr )) {				\
		val = val * 10 + *HeaderStr - '0';			\
		HeaderStr++; HeaderLen++;				\
	}								\
	if ( val > MAX_HEADER_VALUE )					\
		REPORT_ERROR( interp,  "Header value too large", NULL );\
}
/*}}}*/

/*{{{  CHECK_LINK_NUMBER*/
/*	Check if string points to a valid link number
*/
#define	CHECK_LINK_NUMBER( string, interp ){				\
	char *LinkStr = string;						\
	if (( *LinkStr == 'L' )||( *LinkStr == 'l' )){			\
		LinkStr++;						\
		if (( *LinkStr < '0' )||( *LinkStr > '3' ))		\
			REPORT_ERROR( interp,"Expected link number, got ", \
				string );				\
		LinkStr++;						\
		if ( !END_OF_ARG( LinkStr ))				\
			REPORT_ERROR( interp,"Expected link number, got ", \
				string );				\
	}								\
	else								\
		REPORT_ERROR( interp,  "Expected link number, got ", 	\
			string );					\
}
/*}}}*/

/*{{{  CHECK_HEADER*/
/*	Check if string points to a valid destination header
	Flag indicates if the link number should be present
*/
#define CHECK_HEADER( string, LinkOptionalFlag, interp ){		\
	char *tempStr = string;						\
	int len;							\
	if ( !LinkOptionalFlag ){					\
		char temp[3];               				\
		char *tempPtr;						\
		strncpy( temp, tempStr, 2 ); temp[2] = '\0';		\
		tempPtr = temp;						\
		CHECK_LINK_NUMBER( tempPtr, interp );			\
	}								\
	SKIP_LINK_NUMBER( tempStr );					\
	CHECK_ONE_HEADER( tempStr, len, interp );			\
	tempStr += len;							\
	while ( *tempStr == ADDRESS_SEPARATOR ){			\
		tempStr++;						\
		CHECK_ONE_HEADER( tempStr, len, interp );		\
		tempStr += len;						\
	}								\
	if (!END_OF_ARG( tempStr ))					\
		REPORT_ERROR( interp,  "Expected virtual channel header, got ",\
			string );					\
}
/*}}}*/

/*{{{  CHECK_VCHAN*/
/*	Check if string points to a valid virtual channel address
	Flag indicates if the link number should be present
*/
#define	CHECK_VCHAN( string, LinkOptionalFlag, interp )	{		\
	char *ChanStr = string;						\
	int len;							\
	if ( !LinkOptionalFlag ){					\
		char temp[3];               				\
		char *tempPtr;						\
		strncpy( temp, ChanStr, 2 ); temp[2] = '\0';		\
		tempPtr = temp;						\
		CHECK_LINK_NUMBER( tempPtr, interp );			\
	}								\
	SKIP_LINK_NUMBER( ChanStr );					\
	CHECK_ONE_HEADER( ChanStr, len, interp );			\
	ChanStr += len;							\
	if (!END_OF_ARG( ChanStr ))					\
		REPORT_ERROR( interp,  "Expected virtual channel header, got ",\
			string );					\
}
/*}}}*/

/*{{{  CHECK_BINARY_DATA*/
/*	Check if string is of the form <HexByte> <HexByte>...
*/
#if 0
#define CHECK_BINARY_DATA( string, interp ) {				\
	char *DataStr = string;						\
	int val = 0;							\
	while (!END_OF_ARG(DataStr)) {					\
		if ( *DataStr != DATA_SEPARATOR )			\
			val = val * 16 + HEXVAL( *DataStr );		\
		else if ( val > 255 )					\
			REPORT_ERROR( interp,"Expected data in hex, got ", \
				string )				\
		else val = 0;						\
		DataStr++;						\
	}								\
}
#else
#define CHECK_BINARY_DATA( string, interp ) {					\
	char *DataStr = string;							\
	int i, radix;								\
	unsigned long maxelem, val;						\
	char element[MAX_MESSAGE_ELEMENT_LEN];					\
	char ErrorStr[50];							\
	while (!END_OF_ARG(DataStr)){						\
		i = 0;								\
		if (( *DataStr == '0' )&&( *(DataStr+1) == 'x' ))		\
			radix = 16, DataStr += 2;				\
		else if ( *DataStr == '0' )					\
			radix = 8;						\
		else								\
			radix = 10;						\
		while (( !END_OF_ARG(DataStr))&&( *DataStr != DATA_SEPARATOR )){\
			if ( i >= MAX_MESSAGE_ELEMENT_LEN )			\
				REPORT_ERROR( interp,"Err1:Expected data, got ",\
					string )				\
			if (( *(DataStr+1) == DATA_SEPARATOR )||( END_OF_ARG((DataStr+1)))){\
				if ( *DataStr == LONG_WORD_SPEC ){ 		\
					maxelem = 0xffffffff; DataStr++;break; 	\
				}						\
				else if ( *DataStr == SHORT_WORD_SPEC ){	\
					maxelem = 0xffff; DataStr++;break; 	\
				}						\
				else maxelem = 0xff;				\
			}							\
			if ( (HEXVAL( *DataStr )) >= radix ){			\
				sprintf( ErrorStr, "%c, radix %d val %d", *DataStr, radix, HEXVAL(*DataStr) );		\
				REPORT_ERROR( interp,"Err2:Expected data, got ",\
					ErrorStr )				\
			}							\
			element[i++] = *DataStr++;				\
		}								\
		if ( *DataStr == DATA_SEPARATOR ) DataStr++;			\
		element[i] = '\0';						\
		val = StrToNum( element, radix );				\
		if ( val > maxelem )						\
			REPORT_ERROR( interp,"Err3:Expected data, got ", 	\
					 string )				\
	}									\
}
#endif
/*}}}*/

/*{{{  CHECK_HEXVAL*/
/*	Check if string represents a valid 32 bit hex number
*/
#define CHECK_HEXVAL( string, interp ) {				\
	int index = 0;							\
	while (!END_OF_ARG(string+index)){ 				\
		if (!(IS_HEXVAL(string[index] )))			\
			REPORT_ERROR( interp,  "Expected hex val, got ",\
				 string);				\
		index++;						\
	}								\
	if ( index > 8 )REPORT_ERROR( interp, "Expected hex val, got ", \
			string);					\
}
/*}}}*/

/*{{{  GET_HEXVAL*/
/*	Get value of hexadecimal Number
*/
#define GET_HEXVAL( string, Value ) {					\
	char *DataStr = string;						\
	Value = 0;							\
	while (!END_OF_ARG(DataStr)) {					\
		Value = Value * 16 + HEXVAL( *DataStr );		\
		DataStr++;						\
	}								\
}
/*}}}*/
			
