/*	Type definitions for VCP
*/

#ifndef VCP_VCPTYPE_H
#define VCP_VCPTYPE_H

#include "vcpdefs.h"
#include "systype.h"
#include "extype.h"

/*{{{  VIRTUAL_CHAN*/
/*	Virtual channel address
*/
typedef struct {
	int HeaderSize;			/* Length of header in bytes ( =1 or 2 )*/
	long HeaderValue;		/* Value of header			*/
	} VIRTUAL_CHAN;
/*}}}*/

/*{{{  HEADER*/
/*	Header structure. 
	SendHeader is of the form <header 1> ... <header n><VirtualChannel>
*/
typedef struct {
	int  HeaderSize;	/* Size of header 		*/
	char *HeaderPtr;	/* Pointer to actual header 	*/
}HEADER;
/*}}}*/

/*{{{  WAIT_STRUCT*/
/*	Structure on which a process goes to sleep
	SleepFlag is set when the process is asleep
*/
typedef struct {
	Boolean 	SleepFlag;		/* Is the process asleep? */
	WDESC_TYPE 	ProcessDescriptor;	/* Descriptor used to wakeup process */
}WAIT_STRUCT;
/*}}}*/

/*{{{  SEND_REQUEST*/
/*	Queue of messages to be sent on a physical link
	NOTE : qlink should be the first element of the structure
*/
typedef struct send_request {
	QUEUE_CHAIN 	qlink;		/* Pointer to next request 		*/
	struct virtual_chan_info *VirtualChanInfo;	/* Header info for channel	*/
	int		BufferSize;	/* No. of bytes to send			*/
	unsigned char	*BufferPtr;	/* Data to send				*/
	WDESC_TYPE	SenderProcess;	/* Process that is sending the message	*/
	Boolean		EOMFlag;	/* Is this the last packet of msg ?	*/
	int		TimeOut;	/* Max number of microseconds for response	*/
	struct receive_request *TimerRequest;	/* Timer request, if timeout is specified */
	ERROR_TYPE	Error;		/* Error number, VCP_OK if successful	*/
	}SEND_REQUEST;
/*}}}*/

/*{{{  ACK_TO_BE_SENT*/
/*	Queue of Acknowledgements that should be sent
	NOTE : qlink should be the first element of the structure
*/
typedef struct ack_to_be_sent {
	QUEUE_CHAIN	qlink;		/* Pointer to next ACK struct		*/
	int		TimeOut;	/* Timeout in microseconds		*/
	struct receive_request *ReceiveRequest;	/* Request for which ack is to be sent	*/
}ACK_TO_BE_SENT;
/*}}}*/

/*{{{  RECEIVE_REQUEST*/
/*	Queue of posted receive requests
	NOTE : qlink should be the first element of the structure
*/
typedef struct receive_request {
	QUEUE_CHAIN	qlink;		/* Next receive request on this vchan	*/
	struct virtual_chan_info *VirtualChanInfo;	/* Header info for channel	*/
	int		BufferSize;	/* Expected message size		*/
	unsigned char	*BufferPtr;	/* Receive buffer			*/
	Boolean		SleepFlag;	/* Is the receiver process asleep? 	*/
	WDESC_TYPE	ReceiverProcess;/* Process id of receiver		*/
	int		MessageLength;	/* Actual size of message received	*/
	Boolean		EOMFlag;	/* Whether EndOfMessage received	*/
	int		TimeOut;	/* Max number of microseconds for response	*/
	struct receive_request *TimerRequest;	/* Timer request, if timeout is specified */

	ERROR_TYPE	Error;		/* Error number, VCP_OK if successful	*/
}RECEIVE_REQUEST;
/*}}}*/


/*{{{  TIMER_REQUEST*/
/*	Queue of timer requests
	NOTE : qlink should be the first element of the structure
*/
typedef struct timer_request {
	QUEUE_CHAIN	qlink;		/* Next timer request			*/
	int		DeadLine;	/* Deadline for this request		*/
	Boolean		IsSendReq;	/* Whether this is send request		*/
	union {
		SEND_REQUEST *SendRequest;	/* pointer to send request 	*/
		RECEIVE_REQUEST *ReceiveRequest;/* pointer to receive request 	*/
	}request;
}TIMER_REQUEST;
/*}}}*/

/*{{{  ACK_EXPECTEED*/
/*	Acknowledgement expected for send 
*/
typedef struct {
	Boolean		Flag;		/* ACK expected ?			*/
	SEND_REQUEST	*SendRequest;	/* Send for which ACK is expected	*/
}ACK_EXPECTED;
/*}}}*/

/*{{{  VIRTUAL_CHAN_INFO*/
/*	Table containing information on virtual channels currently in use.
	Table is sorted according to the ReceiveHeader
*/
typedef struct virtual_chan_info {
	TABLE_ELEMENT	table;		/* Next elements in table		*/
	VIRTUAL_CHAN	ReceiveHeader;	/* Virtual channel address		*/
	HEADER 		SendHeader;	/* Source of message received		*/
	SEND_REQUEST 	*PendingSendRequests;/* Queued up send requests		*/
	RECEIVE_REQUEST *ReceiveRequest;/* Posted receive requests		*/
	ACK_EXPECTED	AckExpected;	/* ACK due for any send on this vchan?  */
	Boolean		AckNotYetSent;	/* Message arrived, not claimed yet	*/
	Boolean		IgnoreFlag;
}VIRTUAL_CHAN_INFO;
/*}}}*/

/*{{{  PACKETS_RECEIVED*/
/*	Table of unsolicited messages. Table is sorted according to 
	PacketBuffer[1]
*/
typedef struct packets_received {
	TABLE_ELEMENT	table;		/* Next elements in table		*/
	unsigned char	PacketBuffer[MAX_PACKET_SIZE+1]; /* Actual data received	*/
	ERROR_TYPE	Error;		/* Any error encountered ?		*/
}PACKETS_RECEIVED;
/*}}}*/

#endif
